####
#
# fiction-dl
# Copyright (C) (2020 - 2021) Benedykt Synakiewicz <dreamcobbler@outlook.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####

#
#
#
# Imports.
#
#
#

# Application.

from fiction_dl.Concepts.Chapter import Chapter
from fiction_dl.Concepts.Extractor import Extractor
from fiction_dl.Concepts.Story import Story
from fiction_dl.Utilities.HTML import StripHTML

# Standard packages.

from datetime import datetime
import logging
import re
from typing import List, Optional

# Non-standard packages.

from bs4 import BeautifulSoup
from dreamy_utilities.Text import Stringify
from dreamy_utilities.Web import GetHostname

###
#
#
# Extractor designed to work for literotica.com.
#
#
###

class ExtractorLiterotica(Extractor):

    def __init__(self) -> None:

        ##
        #
        # The constructor.
        #
        ##

        super().__init__()

    def GetSupportedHostnames(self) -> List[str]:

        ##
        #
        # Returns a list of hostnames supposed to be supported by the extractor.
        #
        # @return A list of supported hostnames.
        #
        ##

        return [
            "literotica.com"
        ]

    def ScanChannel(self, URL: str) -> Optional[List[str]]:

        ##
        #
        # Scans the channel: generates the list of story URLs.
        #
        # @return **None** when the scan fails, a list of story URLs when it doesn't fail.
        #
        ##

        if (not URL) or (GetHostname(URL) not in self.GetSupportedHostnames()):
            return None

        # Download author's profile page.

        userIDMatch = re.search("\?uid\=(\d+)", URL)
        if not userIDMatch:
            return None

        userID = userIDMatch.group(1)
        userPageURL = f"{self.MEMBER_PAGE_URL}uid={userID}&page=submissions"

        soup = self._webSession.GetSoup(userPageURL)
        if not soup:
            return None

        # Locate all the stories.

        storyURLs = []

        storyHeaderElement = soup.select_one("tr.st-top")
        if not storyHeaderElement:
            return None

        storyRowElement = storyHeaderElement.next_sibling
        while storyRowElement:

            if not storyRowElement.has_attr("class"):
                break

            if "root-story" in storyRowElement["class"]:

                anchorElement = storyRowElement.select_one("a")
                if (not anchorElement) or (not anchorElement.has_attr("href")):
                    continue

                storyURLs.append(anchorElement["href"])
                storyRowElement = storyRowElement.next_sibling

            elif "ser-ttl" in storyRowElement["class"]:

                storyRowElement = storyRowElement.next_sibling
                if (not storyRowElement.has_attr("class")) or ("sl" not in storyRowElement["class"]):
                    continue

                anchorElement = storyRowElement.select_one("a")
                if (not anchorElement) or (not anchorElement.has_attr("href")):
                    continue

                storyURLs.append(anchorElement["href"])
                storyRowElement = storyRowElement.next_sibling

            elif "sl" in storyRowElement["class"]:

                storyRowElement = storyRowElement.next_sibling

            else:

                break

        # Return.

        return storyURLs

    def _InternallyScanStory(
        self,
        URL: str,
        soup: Optional[BeautifulSoup]
    ) -> bool:

        ##
        #
        # Scans the story: generates the list of chapter URLs and retrieves the
        # metadata.
        #
        # @param URL  The URL of the story.
        # @param soup The tag soup.
        #
        # @return **False** when the scan fails, **True** when it doesn't fail.
        #
        ##

        # Extract basic metadata.

        titleElement = soup.select_one("div.b-story-header > h1") or soup.select_one("h1.headline")
        if not titleElement:
            logging.error("Title element not found.")
            return False

        authorElement = soup.select_one("div.b-story-header > span.b-story-user-y > a") or soup.select_one("div.panel > div.y_eS > a.y_eU")
        if not authorElement:
            logging.error("Author element not found.")
            return False

        # Download the author's page.

        if not authorElement.has_attr("href"):
            logging.error("Can't find the URL of the author's page.")
            return False

        authorsPageURL = authorElement["href"]
        authorsPageSoup = self._webSession.GetSoup(authorsPageURL)
        if not authorsPageSoup:
            logging.error(f'Failed to download page: "{authorsPageURL}".')
            return False

        # Extract remaining metadata.

        storyRowElement = None

        for storyLinkElement in authorsPageSoup.select("td.fc > a"):
            if storyLinkElement.get_text().strip() == titleElement.get_text().strip():
                storyRowElement = storyLinkElement.parent.parent
                break

        if not storyRowElement:
            logging.error("Failed to find the story's entry on the author's page.")
            return False

        storyMetadataElements = storyRowElement.find_all("td")
        if len(storyMetadataElements) < 4:
            logging.error("Can't extract metadata from the author's page.")
            return False

        summaryElement = storyMetadataElements[1]
        publishedElement = storyMetadataElements[3]

        # Prepare metadata.

        title = titleElement.get_text().strip()
        datePublished = self._ReformatDate(publishedElement.get_text().strip())
        dateUpdated = self._ReformatDate(publishedElement.get_text().strip())

        # Check if the story belongs to a series.

        seriesRowElement = None

        if storyRowElement.has_attr("class") and ("sl" in storyRowElement["class"]):

            seriesRowElement = storyRowElement.find_previous_sibling(
                "tr",
                {"class": "ser-ttl"}
            )

        if seriesRowElement:

            title = seriesRowElement.get_text().strip()
            chapterDates = []

            seriesChapterRowElement = seriesRowElement.next_sibling
            while seriesChapterRowElement:

                if (not seriesChapterRowElement.has_attr("class")) or ("sl" not in seriesChapterRowElement["class"]):
                    break

                seriesChapterAnchorElement = seriesChapterRowElement.select_one("a")
                if (not seriesChapterAnchorElement) or (not seriesChapterAnchorElement.has_attr("href")):
                    break

                seriesChapterCellElements = seriesChapterRowElement.select("td")
                if seriesChapterCellElements:
                    chapterDates.append(seriesChapterCellElements[-1].get_text().strip())

                self._chapterURLs.append(seriesChapterAnchorElement["href"])
                seriesChapterRowElement = seriesChapterRowElement.next_sibling

            datePublished = self._ReformatDate(chapterDates[0])
            dateUpdated = self._ReformatDate(chapterDates[-1])

        else:

            self._chapterURLs = [self.Story.Metadata.URL]

        # Set the metadata.

        self.Story.Metadata.Title = title
        self.Story.Metadata.Author = authorElement.get_text().strip()

        self.Story.Metadata.DatePublished = datePublished
        self.Story.Metadata.DateUpdated = dateUpdated

        self.Story.Metadata.ChapterCount = len(self._chapterURLs)
        self.Story.Metadata.WordCount = 0

        self.Story.Metadata.Summary = StripHTML(summaryElement.get_text()).strip()

        # Return.

        return True

    def _InternallyExtractChapter(
        self,
        URL: str,
        soup: Optional[BeautifulSoup]
    ) -> Optional[Chapter]:

        ##
        #
        # Extracts specific chapter.
        #
        # @param URL  The URL of the page containing the chapter.
        # @param soup The tag soup of the page containing the chapter.
        #
        # @return **True** if the chapter is extracted correctly, **False** otherwise.
        #
        ##

        # Find the page count of the story.

        pageCount = 1

        if (pageSelectElement := soup.find("select", {"name": "page"})):

            pageCount = len(pageSelectElement.find_all("option"))

            if pageCount < 1:

                logging.error("Failed to read the story's page count.")
                return None

        if (pageSelectElement := soup.select_one("div.panel.clearfix.l_bH")):

            if (linkElements := pageSelectElement.find_all("a")):

                pageCount = int(Stringify(linkElements[-1].encode_contents()))

        # Iterate over pages and read their content.

        content = ""

        for pageIndex in range(1, pageCount + 1):

            pageURL = URL + f"?page={pageIndex}"

            soup = self._webSession.GetSoup(pageURL)
            if not soup:
                logging.error(f'Failed to download page: "{pageURL}".')
                return None

            contentElement = soup.select_one("div.b-story-body-x > div") or soup.select_one("div.panel.article")
            if not contentElement:
                logging.error("Story content element not found.")
                return None

            content += "<br/><br/>" + Stringify(contentElement.encode_contents())

        # Return.

        return Chapter(
            title = None,
            content = content
        )

    def _GetNormalizedStoryURL(self, URL: str) -> str:

        ##
        #
        # Returns a normalized story URL, i.e. one that can be used for anything.
        #
        # @param URL Input URL (given by the user).
        #
        # @return Normalized URL.
        #
        ##

        if not URL:
            return URL

        argumentPosition = URL.find("?")
        if -1 != argumentPosition:
            URL = URL[:argumentPosition]

        return URL

    @staticmethod
    def _ReformatDate(date: str) -> Optional[str]:

        ##
        #
        # Reformats date to ISO 8601.
        #
        # @param date The date in Literotica format (mm/dd/yy).
        #
        # @return The date in ISO 8601 format.
        #
        ##

        if not date:
            return None

        return datetime.strptime(date, "%m/%d/%y").strftime("%Y-%m-%d")

    MEMBER_PAGE_URL = "https://www.literotica.com/stories/memberpage.php?"


####
#
# fiction-dl
# Copyright (C) (2020 - 2021) Benedykt Synakiewicz <dreamcobbler@outlook.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####

#
#
#
# Imports.
#
#
#

# Standard packages.

from typing import List, Optional

#
#
#
# Functions.
#
#
#

def ReadString(
    description: str,
    options: Optional[List[str]] = None,
    default: Optional[str] = None
) -> str:

    ##
    #
    # Reads a string value from the uer.
    #
    # @param description The question asked.
    # @param default     The default value.
    #
    # @return Read value.
    #
    ##

    if options:

        optionsDescription = "/".join(options)
        readValue = input(f"{description} ({optionsDescription}) [{default}]: ")

        return readValue if (readValue in options) else default

    else:

        readValue = input(f"{description}: ")

        return readValue


####
#
# fiction-dl
# Copyright (C) (2020 - 2021) Benedykt Synakiewicz <dreamcobbler@outlook.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####

#
#
#
# Imports.
#
#
#

# Application.

from fiction_dl.Concepts.Processor import Processor

# Standard packages.

import re
from typing import Optional

# Non-standard packages.

from bs4 import BeautifulSoup

#
#
#
# Classes.
#
#
#

##
#
# Clears/fixes the typography.
#
##

class TypographyProcessor(Processor):

    def Process(self, content: str) -> Optional[str]:

        ##
        #
        # Processes the content given and returns a modified version of it.
        #
        # @param content The content to be processed.
        #
        # @return The processed content.
        #
        ##

        if not content:
            return None

        content = self._FixQuotationMarks(content)
        if content is None:
            return None

        content = self._FixPrimitivePunctuation(content)
        if content is None:
            return None

        content = self._FixPunctuationWhitespace(content)
        if content is None:
            return None

        content = self._FixParagraphWhitespace(content)
        if content is None:
            return None

        content = self._ReplacePseudolinesWithLines(content)
        if content is None:
            return None

        return content

    @staticmethod
    def _FixQuotationMarks(content: str) -> Optional[str]:

        ##
        #
        # Replaces quotation marks with prettier ones.
        #
        # @param content Input content.
        #
        # @return Processed content, optionally **None**.
        #
        ##

        if not content:
            return None

        quotationMarkCount = 0;
        processedContent = ""

        for letter in content:

            if '"' != letter:

                processedContent += letter

            else:

                quotationMarkCount += 1

                processedContent += "“" if (quotationMarkCount % 2) else "”"

        return processedContent

    @staticmethod
    def _FixPrimitivePunctuation(content: str) -> Optional[str]:

        ##
        #
        # Fixes "primitive" whitespace. That is, it replaces things like "..." with things like
        # "…".
        #
        # @param content Input content.
        #
        # @return Processed content, optionally **None**.
        #
        ##

        if not content:
            return None

        # Example: ".." -> "…"
        content = re.sub("(\.){2,}", "…", content)

        # Example: ". . ." -> "…"
        content = re.sub("(\.\\s){2,}", "…", content)

        # Example: "………" -> "…"
        content = re.sub("…+", "…", content)

        # Example: "…." -> "…"
        content = re.sub("…\.+", "…", content)

        # Example: "???" -> "?"
        content = re.sub("\?+", "?", content)

        # Example: " - " -> " — "
        content = re.sub("(\\s)-(\\s)", r"\1—\2", content)

        # Example: "-----" -> "—"
        content = re.sub("(\-){2,}", "—", content)

        return content

    @staticmethod
    def _FixPunctuationWhitespace(content: str) -> Optional[str]:

        ##
        #
        # Fixes whitespace around punctuation.
        #
        # @param content Input content.
        #
        # @return Processed content, optionally **None**.
        #
        ##

        if not content:
            return None

        # Example: "  " -> " "
        content = re.sub("(\\s){2,}", " ", content)

        # Example: "A , B" -> "A, B"
        content = re.sub("(\\s),(\\s)", r",\2", content)

        # Example: "A ,B" -> "A, B"
        content = re.sub("(\\s),(\\S)", r", \2", content)

        # Example: "A…B" -> "A… B"
        content = re.sub("(\\S)…(\\S)", r"\1… \2", content)

        # Example: "A … B" -> "A… B"
        content = re.sub("(\\s)…(\\s)", r"…\2", content)

        # Example: " ?" -> "?"
        content = re.sub("(\\s)\\?", "?", content)

        # Example: " !" -> "!"
        content = re.sub("(\\s)\\!", "!", content)

        # Example: "???" -> "?"
        content = re.sub("\?+", "?", content)

        # Example: "A- " -> "A— "
        content = re.sub("(\\S)-(\\s)", r"\1—\2", content)

        # Example: " -A" -> " — A"
        content = re.sub("(\\s)-(\\S)", r" — \2", content)

        # Example: "A—B" -> "A — B"
        content = re.sub("(\\S)—(\\S)", r"\1 — \2", content)

        return content

    @staticmethod
    def _FixParagraphWhitespace(content: str) -> Optional[str]:

        ##
        #
        # Strips preceding and trailing whitespace from paragraph tags.
        #
        # @param content Input content.
        #
        # @return Processed content, optionally **None**.
        #
        ##

        if not content:
            return None

        content = re.sub(r"<p>\s*(.*)\s*</p>", r"<p>\1</p>", content)

        return content

    @staticmethod
    def _ReplacePseudolinesWithLines(content: str) -> Optional[str]:

        ##
        #
        # Replaces pseudolines (i.e. things like "*************" or "----") with proper horizontal
        # lines.
        #
        # @param content Input content.
        #
        # @return Processed content, optionally **None**.
        #
        ##

        if not content:
            return None

        # Replace pseudolines stretching entire paragraphs.

        soup = BeautifulSoup(content, features = "html.parser")

        MAXIMUM_PSEUDOLINE_LENGTH = 30
        for tag in soup.find_all(recursive = True):

            if "p" != tag.name:
                continue

            tagText = tag.get_text()

            if len(tagText) > MAXIMUM_PSEUDOLINE_LENGTH:
                continue

            elif not any(x.isalnum() for x in tagText):
                tag.replace_with(soup.new_tag("hr"))

        content = str(soup)

        # Replace inline pseudolines.

        Pseudolines = [

            "(o\-){3,}",
            "(o \-){3,}",

            "(\*){3,}",
            "(\* ){2,}",

            "(\-){3,}",
            "(\- ){3,}",

            "(\-[a-zA-Z]){3,}",
            "(\- [a-zA-Z]){3,}",

            "(\—){3,}",
            "(\— ){3,}",

            "(\—[a-zA-Z]){3,}",
            "(\— [a-zA-Z]){3,}",

            "(\+){3,}",
            "(\+ ){2,}",

            "(\_){3,}",
            "(\_ ){2,}",

        ]

        for pseudoline in Pseudolines:
            content = re.sub(pseudoline, "<hr/>", content)

        # Remove multiple lines in sequence.

        soup = BeautifulSoup(content, features = "html.parser")

        for tag in soup.find_all(recursive = True):

            if "hr" != tag.name:
                continue

            elif not tag.next_sibling:
                continue

            elif "hr" != tag.next_sibling.name:
                continue

            tag.decompose()

        content = str(soup)

        # Return.

        return content


####
#
# fiction-dl
# Copyright (C) (2020 - 2021) Benedykt Synakiewicz <dreamcobbler@outlook.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####

#
#
#
# Imports.
#
#
#

# Application.

from fiction_dl.Concepts.Chapter import Chapter
from fiction_dl.Concepts.Extractor import Extractor

# Standard packages.

import logging
import re
from typing import List, Optional

# Non-standard packages.

from bs4 import BeautifulSoup
from dreamy_utilities.HTML import ReadElementText
from dreamy_utilities.Interface import Interface
from dreamy_utilities.Text import DeprettifyAmount, DeprettifyNumber, Stringify
from dreamy_utilities.Web import GetHostname

#
#
#
# The class definition.
#
#
#

class ExtractorAO3(Extractor):

    def __init__(self) -> None:

        ##
        #
        # The constructor.
        #
        ##

        super().__init__()

        self._storySoup = None

    def SupportsAuthentication(self) -> bool:

        ##
        #
        # Checks whether the extractor supports user authentication.
        #
        # @return **True** if the site *does* support authentication, **False** otherwise.
        #
        ##

        return True

    def Authenticate(self, interface: Interface) -> bool:

        ##
        #
        # Logs the user in, interactively.
        #
        # @param interface The user interface to be used.
        #
        # @return **True** if the user has been authenticated correctly, **False** otherwise.
        #
        ##

        # Download the log-in page.

        soup = self._webSession.GetSoup(self._LOGIN_URL)
        if not soup:
            logging.error("Failed to download the log-in page.")
            return self.AuthenticationResult.FAILURE

        formElement = soup.select_one("form#new_user_session_small")
        if not formElement:
            logging.error("Log-in form element not found.")
            return self.AuthenticationResult.FAILURE

        authenticityTokenElement = formElement.find("input", {"name": "authenticity_token"})
        if not authenticityTokenElement:
            logging.error("\"authenticity_token\" input field not found.")
            return self.AuthenticationResult.FAILURE
        elif not authenticityTokenElement.has_attr("value"):
            logging.error("\"authenticity_token\" input field doesn't have a value.")
            return self.AuthenticationResult.FAILURE

        authenticityToken = authenticityTokenElement["value"].strip()

        # Read the username and the password.

        userName = ""
        password = ""

        if ExtractorAO3._userName is None:

            interface.GrabUserAttention()

            userName = interface.ReadString("Your username")
            if userName:
                password = interface.ReadPassword("Your password")

        else:

            userName = ExtractorAO3._userName
            password = ExtractorAO3._userPassword

        if not userName:
            userName = ""

        # Remember the username and the password.

        ExtractorAO3._userName = userName
        ExtractorAO3._userPassword = password

        # Decide whether to log-in.

        if (not userName) or (not password):
            return self.AuthenticationResult.ABANDONED

        # Attempt to log-in.

        data = {
            "user[login]": userName,
            "user[password]": password,
            "user[remember_me]": "1",
            "authenticity_token": authenticityToken,
        }

        response = self._webSession.Post(
            URL = self._LOGIN_URL,
            payload = data
        )

        # Verify the response and return.

        if (not response) or ("doesn't match our records" in response.lower()):
            return self.AuthenticationResult.FAILURE

        return self.AuthenticationResult.SUCCESS

    def GetSupportedHostnames(self) -> List[str]:

        ##
        #
        # Returns a list of hostnames supposed to be supported by the extractor.
        #
        # @return A list of supported hostnames.
        #
        ##

        return [
            "archiveofourown.org"
        ]

    def RequiresBreaksBetweenRequests(self) -> bool:

        ##
        #
        # Does the extractor require the application to sleep between subsequent reqests?
        #
        # @return **True** if it does, **False** otherwise.
        #
        ##

        return False

    def ScanChannel(self, URL: str) -> Optional[List[str]]:

        ##
        #
        # Scans the channel: generates the list of story URLs.
        #
        # @return **None** when the scan fails, a list of story URLs when it doesn't fail.
        #
        ##

        if (not URL) or (GetHostname(URL) not in self.GetSupportedHostnames()):
            return None

        seriesMatch = re.search("/series/([a-zA-Z0-9_]+)", URL)
        if seriesMatch:
            return self._ScanWorks(f"{self._BASE_SERIES_URL}/{seriesMatch.group(1)}")

        collectionMatch = re.search("/collections/([a-zA-Z0-9_]+)", URL)
        if collectionMatch:
            return self._ScanWorks(f"{self._BASE_COLLECTION_URL}/{collectionMatch.group(1)}/works")

        usernameMatch = re.search("/users/([a-zA-Z0-9_]+)", URL)
        if usernameMatch:
            return self._ScanWorks(f"{self._BASE_USER_URL}/{usernameMatch.group(1)}/works")

        return None

    def _InternallyScanStory(
        self,
        URL: str,
        soup: Optional[BeautifulSoup]
    ) -> bool:

        ##
        #
        # Scans the story: generates the list of chapter URLs and retrieves the
        # metadata.
        #
        # @param URL  The URL of the story.
        # @param soup The tag soup.
        #
        # @return **False** when the scan fails, **True** when it doesn't fail.
        #
        ##

        # Store the tag soup.

        self._storySoup = soup

        # Extract the metadata.

        self.Story.Metadata.Title = ReadElementText(self._storySoup, "h2.title")
        self.Story.Metadata.Author = ReadElementText(self._storySoup, "a[rel~=author]") or "Anonymous"

        self.Story.Metadata.DatePublished = ReadElementText(self._storySoup, "dd.published")
        self.Story.Metadata.DateUpdated = ReadElementText(self._storySoup, "dd.status") or self.Story.Metadata.DatePublished

        chapterCount = DeprettifyAmount(ReadElementText(self._storySoup, "dd.chapters"))[0]

        self.Story.Metadata.ChapterCount = DeprettifyAmount(ReadElementText(self._storySoup, "dd.chapters"))[0]
        self.Story.Metadata.WordCount = DeprettifyNumber(ReadElementText(self._storySoup, "dd.words"))

        self.Story.Metadata.Summary = ReadElementText(self._storySoup, "blockquote.userstuff") or "No summary."

        # Validate story metadata.

        if (missingMetadata := self.Story.Metadata.AreValuesMissing()):

            logging.error(f"Failed to read following story metadata: {', '.join(missingMetadata)}.")

            return False

        # Return.

        return True

    def ExtractChapter(self, index: int) -> Optional[Chapter]:

        ##
        #
        # Extracts specific chapter.
        #
        # @param index The index of the chapter to be extracted.
        #
        # @return **True** if the chapter is extracted correctly, **False** otherwise.
        #
        ##

        if 1 == self.Story.Metadata.ChapterCount:

            titleElement = None

            contentElement = self._storySoup.select_one("div#chapters div.userstuff")
            if not contentElement:
                logging.error("Content element not found.")
                return None

            if (landmarkElement := contentElement.select_one("h3#work")):
                landmarkElement.decompose()

            return Chapter(
                title = titleElement.get_text().strip() if titleElement else None,
                content = Stringify(contentElement.encode_contents())
            )

        else:

            chapterElements = self._storySoup.select("div#chapters > div.chapter")
            if index > len(chapterElements):
                logging.error(
                    f"Trying to extract chapter {index}. "
                    f"Only {len(chapterElements)} chapter(s) located. "
                    f"The story supposedly has {self.Story.Metadata.ChapterCount} chapter(s)."
                )
                return None

            currentChapterElement = chapterElements[index - 1]

            titleElement = currentChapterElement.select_one("h3.title")
            contentElement = currentChapterElement.select_one("div.userstuff")

            if (landmarkElement := contentElement.select_one("h3#work")):
                landmarkElement.decompose()

            return Chapter(
                title = titleElement.get_text().strip() if titleElement else None,
                content = Stringify(contentElement.encode_contents())
            )

    def _ScanWorks(self, URL: str) -> Optional[List[str]]:

        ##
        #
        # Scans a list of works: generates the list of story URLs.
        #
        # @param URL The URL.
        #
        # @return **None** when the scan fails, a list of story URLs when it doesn't fail.
        #
        ##

        # Check the arguments.

        if not URL:
            return None

        # Download page soup.

        soup = self._webSession.GetSoup(URL)
        if not soup:
            return None

        # Locate all the pages of the list.

        pageURLs = []

        if not (paginationElement := soup.find("ol", {"title": "pagination"})):

            pageURLs.append(URL)

        else:

            pageButtonElements = paginationElement.find_all("li")
            pageCount = len(pageButtonElements) - 2

            if pageCount < 1:
                logging.error("Invalid number of pages on the Works webpage.")
                return None

            for index in range(1, pageCount + 1):
                pageURLs.append(f"{URL}?page={index}")

        # Read all the stories on all the pages.

        storyURLs = []

        for pageURL in pageURLs:

            soup = self._webSession.GetSoup(pageURL)
            if not soup:
                logging.error("Failed to download a page of the Works webpage.")
                continue

            for storyElement in soup.select("li.work"):

                linkElement = storyElement.find("a")
                if (not linkElement) or (not linkElement.has_attr("href")):
                    logging.error("Failed to retrieve story URL from the Works webpage.")
                    continue

                storyID = ExtractorAO3._GetStoryID(linkElement["href"])
                if not storyID:
                    logging.error("Failed to retrieve story ID from its URL.")
                    continue

                storyURLs.append(f"{ExtractorAO3._BASE_WORK_URL}/{storyID}")

        # Return.

        return storyURLs

    @staticmethod
    def _GetNormalizedStoryURL(URL: str) -> Optional[str]:

        ##
        #
        # Returns a normalized story URL, i.e. one that can be used for anything.
        #
        # @param URL Input URL.
        #
        # @return Normalized URL.
        #
        ##

        if not URL:
            return None

        return ExtractorAO3._GetAdultFullStoryURL(
            f"{ExtractorAO3._BASE_WORK_URL}/{ExtractorAO3._GetStoryID(URL)}"
        )

    @staticmethod
    def _GetPrettyStoryURL(URL: str) -> Optional[str]:

        ##
        #
        # Returns a "pretty" story URL, i.e. one that can be used for printing in story content.
        #
        # @param URL Input URL (given by the user).
        #
        # @return Pretty URL.
        #
        ##

        if not URL:
            return None

        return f"{ExtractorAO3._BASE_WORK_URL}/{ExtractorAO3._GetStoryID(URL)}"

    @staticmethod
    def _GetAdultFullStoryURL(URL: str) -> Optional[str]:

        ##
        #
        # Returns a URL leading to the page containg the whole story content, with adult-mode
        # enabled.
        #
        # @param URL Story URL.
        #
        # @return The code of the page containing the whole work.
        #
        ##

        if not URL:
            return None

        return URL + "?view_adult=true&view_full_work=true"

    @staticmethod
    def _GetStoryID(URL: str) -> Optional[str]:

        ##
        #
        # Retrieves story ID from story URL.
        #
        # @param URL Story URL.
        #
        # @return Story ID.
        #
        ##

        if not URL:
            return None

        storyIDMatch = re.search("/works/(\d+)", URL)
        if not storyIDMatch:
            return None

        return storyIDMatch.group(1)

    _userName = None
    _userPassword = None

    _LOGIN_URL = "https://archiveofourown.org/users/login"
    _BASE_WORK_URL = "https://archiveofourown.org/works"
    _BASE_USER_URL = "https://archiveofourown.org/users"
    _BASE_SERIES_URL = "https://archiveofourown.org/series"
    _BASE_COLLECTION_URL = "https://archiveofourown.org/collections"


####
#
# fiction-dl
# Copyright (C) (2020 - 2021) Benedykt Synakiewicz <dreamcobbler@outlook.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####

#
#
#
# Imports.
#
#
#

# Application.

from fiction_dl.Concepts.Chapter import Chapter
from fiction_dl.Concepts.Extractor import Extractor
import fiction_dl.Configuration as Configuration

# Standard packages.

from datetime import datetime
import logging
import re
from typing import List, Optional, Tuple

# Non-standard packages.

from bs4 import BeautifulSoup
from dreamy_utilities.Text import PrettifyDate, Stringify
from dreamy_utilities.Web import GetSiteURL

#
#
#
# Classes.
#
#
#

##
#
# An extractor dedicated for Nifty.org.
#
##

class ExtractorNifty(Extractor):

    def __init__(self) -> None:

        ##
        #
        # The constructor.
        #
        ##

        super().__init__()

        self._downloadStorySoupWhenScanning = False
        self._downloadChapterSoupWhenExtracting = False

    def GetSupportedHostnames(self) -> List[str]:

        ##
        #
        # Returns a list of hostnames supposed to be supported by the extractor.
        #
        # @return A list of supported hostnames.
        #
        ##

        return [
            "nifty.org",
        ]

    def _InternallyScanStory(
        self,
        URL: str,
        soup: Optional[BeautifulSoup]
    ) -> bool:

        ##
        #
        # Scans the story: generates the list of chapter URLs and retrieves the
        # metadata.
        #
        # @param URL  The URL of the story.
        # @param soup The tag soup.
        #
        # @return **False** when the scan fails, **True** when it doesn't fail.
        #
        ##

        # Is it a single chapter story?

        pageCode = self._webSession.Get(URL, textEncoding = "ascii")
        if not pageCode:
            logging.error("Failed to download story page when scanning.")
            return False

        isHTMLCode = (-1 != pageCode.find("<html>"))

        # Process a multi-chapter story.

        if isHTMLCode:

            # Create tag soup.

            soup = self._webSession.GetSoup(URL)

            # Create a list of chapters.

            chapterElements = soup.select("table.table.table-xtra-condensed > tr")
            if not chapterElements:
                logging.error("List of chapters not found.")
                return False

            chapterElements = chapterElements[1:]

            # Process chapters.

            for rowElement in chapterElements:

                cellElements = rowElement.select("td")
                if (not cellElements) or (len(cellElements) < 3):
                    continue

                anchorElement = cellElements[2].select_one("a")
                if (not anchorElement) or (not anchorElement.has_attr("href")):
                    continue

                self._chapterURLs.append(self.Story.Metadata.URL + anchorElement["href"])

            if not self._chapterURLs:
                return False

            self._chapterURLs.reverse()

            # Read the first chapter.

            firstChapterText = self._webSession.Get(self._chapterURLs[0], textEncoding = "ascii")
            if not soup:
                logging.error("Failed to download the first chapter.")
                return False

            firstChapterMetadata = self._ReadChapterMetadata(firstChapterText)
            if not firstChapterMetadata:
                logging.error("Failed to read metadata from the first chapter of the story.")
                return False

            # Read the last chapter.

            lastChapterText = self._webSession.Get(self._chapterURLs[-1], textEncoding = "ascii")
            if not soup:
                logging.error("Failed to download the last chapter.")
                return False

            lastChapterMetadata = self._ReadChapterMetadata(lastChapterText)
            if not lastChapterMetadata:
                logging.error("Failed to read metadata from the last chapter of the story.")
                return False

            # Prepare the metadata.

            title = firstChapterMetadata[0]
            author = firstChapterMetadata[1]
            publicationDate = firstChapterMetadata[2]
            updateDate = lastChapterMetadata[2]

            # Set the metadata.

            self.Story.Metadata.Title = title
            self.Story.Metadata.Author = author

            self.Story.Metadata.DatePublished = publicationDate
            self.Story.Metadata.DateUpdated = updateDate

            self.Story.Metadata.ChapterCount = len(self._chapterURLs)
            self.Story.Metadata.WordCount = 0

            self.Story.Metadata.Summary = "No summary."

        # Process a single-chapter story.

        else:

            # Read chapter metadata.

            chapterMetadata =  self._ReadChapterMetadata(pageCode)
            if not chapterMetadata:
                logging.error("Failed to read metadata from the only chapter of the story.")
                return False

            # Set the metadata.

            self.Story.Metadata.Title = chapterMetadata[0]
            self.Story.Metadata.Author = chapterMetadata[1]

            self.Story.Metadata.DatePublished = chapterMetadata[2]
            self.Story.Metadata.DateUpdated = self.Story.Metadata.DatePublished

            self.Story.Metadata.ChapterCount = 1
            self.Story.Metadata.WordCount = 0

            self.Story.Metadata.Summary = "No summary."

            # Create a list of chapter URLs.

            self._chapterURLs = [URL]

        # Return.

        return True

    def _InternallyExtractChapter(
        self,
        URL: str,
        soup: Optional[BeautifulSoup]
    ) -> Optional[Chapter]:

        ##
        #
        # Extracts specific chapter.
        #
        # @param URL  The URL of the page containing the chapter.
        # @param soup The tag soup of the page containing the chapter.
        #
        # @return **True** if the chapter is extracted correctly, **False** otherwise.
        #
        ##

        # Read the chapter.

        chapterText = self._webSession.Get(URL, textEncoding = "ascii")
        if not chapterText:
            logging.error("Failed to download a chapter.")
            return False

        chapterText = chapterText.splitlines()
        if len(chapterText) < 4:
            logging.error("Invalid chapter format.")
            return False

        chapterText = chapterText[3:]

        # Format the content.

        chapterCode = ""
        currentParagraphCode = ""

        for line in chapterText:

            if not line:

                chapterCode += f"<p>{currentParagraphCode}</p>"
                currentParagraphCode = ""

            else:

                currentParagraphCode += f" {line.strip()}"

        # Return.

        return Chapter(content = chapterCode)

    @classmethod
    def _ReadChapterMetadata(cls, text: str) -> Optional[Tuple[str, str, str]]:

        ##
        #
        # Reads chapter metadata from its content.
        #
        # @param text Chapter text.
        #
        # @return A tuple consisting of three values: story title, story author, chapter publication
        #         date; optionally **None**.
        #
        ##

        if not text:
            return None

        # Split the chapter text into lines.

        text = text.splitlines()
        if len(text) < 3:
            return None

        # Retrieve lines containing relevant metadata.

        dateString = text[0]
        authorString = text[1]
        titleString = text[2]

        # Retrieve metadata from appropriate lines.

        dateMatch = re.match("Date: (.+)", dateString)
        if not dateMatch:
            return None

        authorMatch = re.match("From: ([^<]+) <", authorString)
        if not authorMatch:
            return None

        titleMatch = re.match("Subject: (.+)", titleString)
        if not titleMatch:
            return None

        # Process and return the metadata.

        date = datetime.strptime(dateMatch.group(1), cls._DATE_FORMAT).strftime("%Y-%m-%d")
        author = authorMatch.group(1).strip()
        title = titleMatch.group(1).strip()

        return (title, author, date)

    # Class constants.

    _DATE_FORMAT = "%a, %d %b %Y %H:%M:%S %z"



####
#
# fiction-dl
# Copyright (C) (2020 - 2021) Benedykt Synakiewicz <dreamcobbler@outlook.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####

#
#
#
# Imports.
#
#
#

# Application.

from fiction_dl.Concepts.Chapter import Chapter
from fiction_dl.Concepts.Extractor import Extractor
from fiction_dl.Concepts.Story import Story
from fiction_dl.Utilities.HTML import StripHTML

# Standard packages.

import logging
import re
from typing import List, Optional

# Non-standard packages.

from bs4 import BeautifulSoup
from dreamy_utilities.Text import FindFirstMatch, Stringify

#
#
#
# Classes.
#
#
#

##
#
# An extractor dedicated for WuxiaWorld.com.
#
##

class ExtractorWuxiaWorld(Extractor):

    def __init__(self) -> None:

        ##
        #
        # The constructor.
        #
        ##

        super().__init__()

        self._downloadStorySoupWhenScanning = False

    def GetSupportedHostnames(self) -> List[str]:

        ##
        #
        # Returns a list of hostnames supposed to be supported by the extractor.
        #
        # @return A list of supported hostnames.
        #
        ##

        return [
            "wuxiaworld.com",
        ]

    def _InternallyScanStory(
        self,
        URL: str,
        soup: Optional[BeautifulSoup]
    ) -> bool:

        ##
        #
        # Scans the story: generates the list of chapter URLs and retrieves the
        # metadata.
        #
        # @param URL  The URL of the story.
        # @param soup The tag soup.
        #
        # @return **False** when the scan fails, **True** when it doesn't fail.
        #
        ##

        # Generate normalized URL.

        storyIdentifier = FindFirstMatch(URL, "/novel/([a-zA-Z0-9-]+)/")
        if not storyIdentifier:
            logging.error("Failed to retrieve the story identifier.")
            return False

        normalizedURL = self._BASE_NOVEL_URL + storyIdentifier
        soup = self._webSession.GetSoup(normalizedURL)
        if not soup:
            logging.error(f"Failed to download page: \"{normalizedURL}\".")
            return False

        # Locate relevant page elements.

        titleElement = soup.select_one("div.novel-body > h2")
        if not titleElement:
            logging.error("Title element not found.")
            return False

        authorElement = soup.select_one("div.novel-body > div > div > dd")
        if not authorElement:
            logging.error("Author element not found.")
            return False

        summaryElement = soup.select_one("div.novel-bottom > div > div.fr-view")
        if not summaryElement:
            logging.error("Summary element not found.")
            return False

        # Retrieve chapter URLs.

        for element in soup.select("div#chapters a"):

            if (not element.has_attr("href")) or (not element["href"].startswith("/novel/")):
                continue

            self._chapterURLs.append(self._BASE_URL + element["href"])

        # Set the metadata.

        self.Story.Metadata.Title = titleElement.get_text().strip()
        self.Story.Metadata.Author = self._ProcessAuthorName(authorElement.get_text())

        self.Story.Metadata.DatePublished = "?"
        self.Story.Metadata.DateUpdated = "?"

        self.Story.Metadata.ChapterCount = len(self._chapterURLs)
        self.Story.Metadata.WordCount = 0

        self.Story.Metadata.Summary = StripHTML(summaryElement.get_text().strip())

        # Return.

        return True

    def _InternallyExtractChapter(
        self,
        URL: str,
        soup: Optional[BeautifulSoup]
    ) -> Optional[Chapter]:

        ##
        #
        # Extracts specific chapter.
        #
        # @param URL  The URL of the page containing the chapter.
        # @param soup The tag soup of the page containing the chapter.
        #
        # @return **True** if the chapter is extracted correctly, **False** otherwise.
        #
        ##

        # Locate relevant page elements.

        titleElement = soup.select_one("div#chapter-outer > div.caption > div > h4")
        # No error-checking here. Not sure if every chapter has to have a title on WW.

        contentElement = soup.select_one("div#chapter-content")
        if not contentElement:
            logging.error("Content element not found.")
            return False

        # Return.

        return Chapter(
            titleElement.get_text().strip() if titleElement else "",
            Stringify(contentElement.encode_contents())
        )

    @staticmethod
    def _ProcessAuthorName(authorName: str) -> str:

        ##
        #
        # Processes author name (removes unnecessary parts).
        #
        # @param authorName The original author name, as it was retrieved.
        #
        # @return Processed author name.
        #
        ##

        if not authorName:
            return authorName

        position = authorName.find("/")
        if -1 != position:
            authorName = authorName[:position - 1]

        position = authorName.find("(")
        if -1 != position:
            authorName = authorName[:position - 1]

        return authorName.strip()

    _BASE_URL = "https://www.wuxiaworld.com"
    _BASE_NOVEL_URL = "https://www.wuxiaworld.com/novel/"


####
#
# fiction-dl
# Copyright (C) (2020 - 2021) Benedykt Synakiewicz <dreamcobbler@outlook.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####

#
#
#
# Imports.
#
#
#

# Application.

from fiction_dl.Concepts.Story import Story
from fiction_dl.Concepts.StoryPackage import StoryPackage

# Standard packages.

from pathlib import Path
from typing import List, Union

#
#
#
# Classes.
#
#
#

##
#
# Represents an abstract formatter - that is, a thing generating an output file from a Story.
#
##

class Formatter:

    def __init__(self, embedImages: bool = True) -> None:

        ##
        #
        # The constructor.
        #
        # @param embedImages Embed images in the output file.
        #
        ##

        self._embedImages = embedImages

    def FormatAndSave(self, story: Union[Story, StoryPackage], filePath: Path) -> bool:

        ##
        #
        # Formats the story (or the story package) and saves it to the output file.
        #
        # @param story    The story/story package to be formatted and saved.
        # @param filePath The path to the output file.
        #
        # @return **True** if the output file was generated and saved without problems, **False**
        #         otherwise.
        #
        ##

        raise NotImplementedError()


####
#
# fiction-dl
# Copyright (C) (2020 - 2021) Benedykt Synakiewicz <dreamcobbler@outlook.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####

#
#
#
# Imports.
#
#
#

# Application.

from fiction_dl.Extractors.ExtractorXenForo import ExtractorXenForo

# Standard packages.

from typing import List

#
#
#
# Classes.
#
#
#

##
#
# Extractor designed for SufficientVelocity.com. Internally it uses the XenForo extractor.
#
##

class ExtractorSufficientVelocity(ExtractorXenForo):

    def __init__(self) -> None:

        ##
        #
        # The constructor.
        #
        ##

        super().__init__("https://forums.sufficientvelocity.com/")

    def GetSupportedHostnames(self) -> List[str]:

        ##
        #
        # Returns a list of hostnames supposed to be supported by the extractor.
        #
        # @return A list of supported hostnames.
        #
        ##

        return [
            "sufficientvelocity.com"
        ]


####
#
# fiction-dl
# Copyright (C) (2020 - 2021) Benedykt Synakiewicz <dreamcobbler@outlook.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####

#
#
#
# Imports.
#
#
#

# Standard packages.

from pathlib import Path

# Non-standard packages.

import fitz

#
#
#
# Functions.
#
#
#

def RenderPDFPageToBytes(documentFilePath: Path, pageIndex: int) -> bytes:

    ##
    #
    # Renders given page of a PDF document to image, then returns the image data.
    #
    # @param documentFilePath The path to the document in question.
    # @param pageIndex        Index of the page to be rendered.
    #
    # @return Encoded image data, in JPEG format.
    #
    ##

    document = fitz.open(documentFilePath)
    page = document.loadPage(pageIndex)

    return page.getPixmap().getImageData(output = "jpeg")


####
#
# fiction-dl
# Copyright (C) (2020 - 2021) Benedykt Synakiewicz <dreamcobbler@outlook.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####

#
#
#
# Imports.
#
#
#

# Standard packages.

from typing import Optional

# Non-standard packages.

from dreamy_utilities.Text import PrettifyTitle

#
#
#
# Classes.
#
#
#

##
#
# Represents a single chapter of a story.
#
##

class Chapter:

    def __init__(self, title: Optional[str] = None, content: Optional[str] = None) -> None:

        ##
        #
        # The constructor.
        #
        # @param title   The title of the chapter.
        # @param content The content of the chapter.
        #
        ##

        self.Title = title
        self.Content = content

    def Process(self) -> None:

        ##
        #
        # Processes the chapter. To be used before generating output file.
        #
        ##

        self.Title = PrettifyTitle(self.Title, removeContext = True)

    def __bool__(self) -> bool:

        ##
        #
        # The bool operator.
        #
        # @return **True** if the chapter has any content.
        #
        ##

        return bool(self.Content)


####
#
# fiction-dl
# Copyright (C) (2020 - 2021) Benedykt Synakiewicz <dreamcobbler@outlook.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####

#
#
#
# Imports.
#
#
#

# Application.

from fiction_dl.Concepts.Story import Story

# Standard packages.

import re
from typing import Optional

# Non-standard packages.

from dreamy_utilities.Filesystem import GetSanitizedFileName
from dreamy_utilities.Text import IsRomanNumeral, IsStringEmpty, PrettifyTitle
import pykakasi

#
#
#
# Functions.
#
#
#

def GetPrintableStoryTitle(story: Story) -> str:

    prettifiedMetadata = story.Metadata.GetPrettified()

    title = Transliterate(prettifiedMetadata.Title)
    title = GetSanitizedFileName(title)
    title = re.sub("\s+", " ", title)

    return title.strip()

def GetTitleProper(title: str) -> Optional[str]:

    ##
    #
    # Retrieves the proper title of the story (removing things like parts after a semicolon).
    #
    # @param title The full title.
    #
    # @return The title proper.
    #
    ##

    if not title:
        return None

    titleProper = PrettifyTitle(title, removeContext = True)

    if titleProper.endswith("?") or titleProper.endswith(".") or titleProper.endswith("!"):
        titleProper = titleProper[:-1]

    separatorOccurence = titleProper.find(":")
    if -1 != separatorOccurence:
        titleProper = titleProper[:separatorOccurence]

    separatorOccurence = titleProper.find(".")
    if -1 != separatorOccurence:
        titleProper = titleProper[:separatorOccurence]

    separatorOccurence = titleProper.find("?")
    if -1 != separatorOccurence:
        titleProper = titleProper[:separatorOccurence]

    separatorOccurence = titleProper.find("!")
    if -1 != separatorOccurence:
        titleProper = titleProper[:separatorOccurence]

    words = titleProper.split()
    if IsRomanNumeral(words[-1]):
        titleProper = " ".join(words[:-1])

    return titleProper.strip()

def Transliterate(string: str) -> str:

    ##
    #
    # Transliterates given string to ASCII.
    #
    # @param string The input string.
    #
    # @return The transliterated input string.
    #
    ##

    if not string:
        return string

    kakasi = pykakasi.kakasi()
    kakasi.setMode("H", "a")
    kakasi.setMode("K", "a")
    kakasi.setMode("J", "a")
    kakasi.setMode("r", "Hepburn")
    kakasi.setMode("s", True)
    kakasi.setMode("C", True)

    string = kakasi.getConverter().do(string)

    return string


####
#
# fiction-dl
# Copyright (C) (2020 - 2021) Benedykt Synakiewicz <dreamcobbler@outlook.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####

###
#
#
# Imports.
#
#
###

# Application.

import fiction_dl.Configuration as Configuration

# Standard packages.

from setuptools import find_packages, setup

# Non-standard packages.

from dreamy_utilities.Filesystem import ReadTextFile

###
#
#
# The start-up routine.
#
#
###

setup(

    name = Configuration.ApplicationName,
    version = Configuration.ApplicationVersion,

    author = Configuration.CreatorName,
    author_email = Configuration.CreatorContact,

    description = Configuration.ApplicationShortDescription,
    long_description = ReadTextFile("README.md"),
    long_description_content_type = "text/markdown",

    url = Configuration.ApplicationURL,

    install_requires = [
        "PyMuPDF",
        "bleach",
        "bs4",
        "dreamy-utilities>=1.2.0",
        "ebooklib",
        "fake-useragent",
        "html5lib",
        "lxml",
        "markdown",
        "numpy",
        "opencv-python",
        "pillow",
        "praw",
        "pykakasi",
        "pyopenssl",
    ],

    packages = find_packages(),
    package_data = {

        "": [
            "*.css",
            "*.html",
            "*.odt",
            "*.xml"
        ]

    },

    entry_points = {

        "console_scripts": [
            "fiction-dl = fiction_dl.__main__:Main",
            "f-dl = fiction_dl.__main__:Main",
        ]

    },

    classifiers = [

        # Development Status :: 1 - Planning
        # Development Status :: 2 - Pre-Alpha
        # Development Status :: 3 - Alpha
        # Development Status :: 4 - Beta
        # Development Status :: 5 - Production/Stable
        # Development Status :: 6 - Mature
        # Development Status :: 7 - Inactive
        "Development Status :: 5 - Production/Stable",

        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License (GPL)",

        "Operating System :: OS Independent",
        "Environment :: Console",

        "Intended Audience :: End Users/Desktop",
        "Topic :: Internet :: WWW/HTTP",

    ],

    python_requires = ">=3.8",

)


####
#
# fiction-dl
# Copyright (C) (2020 - 2021) Benedykt Synakiewicz <dreamcobbler@outlook.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####

#
#
#
# Imports.
#
#
#

# Application.

from fiction_dl.Concepts.Image import Image

# Standard packages.

from typing import Dict, List, Optional
from urllib.parse import urlparse

# Non-standard packages.

import bleach
from bs4 import BeautifulSoup
from dreamy_utilities.HTML import UnescapeHTMLEntities

#
#
#
# Functions.
#
#
#

def CleanHTML(code: str) -> Optional[str]:

    ##
    #
    # Cleans HTML code.
    #
    # @param code The input code.
    #
    # @return Cleaned input code, or None.
    #
    ##

    if not code:
        return None

    # Unescape entities and remove non-breaking spaces.

    code = UnescapeHTMLEntities(code)
    code = code.replace(u"\u00A0", " ")

    # Return.

    return code

def FindImagesInCode(code: str) -> List[Image]:

    ##
    #
    # Creates a list of Images, based on "img" tags in given code.
    #
    # @param code The input code.
    #
    # @return A List of Image objects, in the same order "img" tags were encountered in the code.
    #
    ##

    soup = BeautifulSoup(code, features = "html.parser")

    return [Image(UnescapeHTMLEntities(tag["src"])) for tag in soup.find_all("img")]

def IsURLAbsolute(URL: str) -> bool:

    ##
    #
    # Checks whether given URL is absolute.
    #
    # @param URL The URL.
    #
    # @return **True** if the URL is absolute, **False** otherwise.
    #
    ##

    if not URL:
        return True

    return bool(urlparse(UnescapeHTMLEntities(URL)).netloc)

def MakeURLAbsolute(URL: str, baseURL: str) -> Optional[str]:

    ##
    #
    # Converts relative URL to absolute URL.
    #
    # @param URL     The input URL.
    # @param baseURL The URL to be prepended in front of the relative input URL.
    #
    # @return Absolute version of the input URL. Optionally None, if something goes wrong.
    #
    ##

    if (not URL) or (not baseURL):
        return None

    elif IsURLAbsolute(URL):
        return URL

    URL = URL if URL.startswith("/") else f"/{URL}"

    return f"{baseURL}{URL}"

def ReformatHTMLToXHTML(code: str) -> Optional[str]:

    ##
    #
    # Reformats given HTML code as valid XHTML.
    #
    # @param code Input HTML code.
    #
    # @return Output XHTML code.
    #
    ##

    if not code:
        return None

    soup = BeautifulSoup(code, "html.parser")
    code = str(soup)

    soup = BeautifulSoup(code, "lxml")
    code = str(soup)

    return code

def StripEmptyTags(
    code: str,
    validEmptyTags: List[str] = [],
    validEmptyTagAttributes: Dict = {}
) -> Optional[str]:

    ##
    #
    # Strips all empty tags from code.
    #
    # @param code                    The input code.
    # @param validEmptyTags          A list of tags allowed to be empty.
    # @param validEmptyTagAttributes A dictionary of attributes with values that allow empty tag to
    #                                get away with being empty.
    #
    # @return The processed code.
    #
    ##

    if not code:
        return None

    soup = BeautifulSoup(code, features = "html.parser")

    tagsStripped = 1

    while tagsStripped:

        tagsStripped = 0

        for tag in soup.find_all(True, recursive = True):

            if tag.name in validEmptyTags:
                continue

            elif tag.find_all(True, recursive = True):
                continue

            elif len(tag.get_text(strip = True)):
                continue

            elif validEmptyTagAttributes:

                validBecauseOfAttributes = False

                for attribute, value in validEmptyTagAttributes.items():

                    if tag.has_attr(attribute) and (value == tag[attribute]):
                        validBecauseOfAttributes = True

                if validBecauseOfAttributes:
                    continue

            tag.decompose()
            tagsStripped += 1

    return str(soup)

def StripHTML(code: str, paragraphSeparator: str = "\n\n") -> Optional[str]:

    ##
    #
    # Converts HTML code to raw text.
    #
    # @param code               The input code.
    # @param paragraphSeparator Text to be inserted in place of paragraph breaks ("</p><p>").
    #
    # @return Raw text generated from the input code.
    #
    ##

    if not code:
        return None

    return BeautifulSoup(
        StripTags(CleanHTML(code), ["p"]),
        "html.parser"
    ).get_text(separator = paragraphSeparator)

def StripTags(code: str, validTags: List[str] = []) -> Optional[str]:

    ##
    #
    # Strips all tags from code, with the exception of those designated "valid" tags.
    #
    # @param code      The input code.
    # @param validTags A List of tags **not** to be stripped from the code.
    #
    # @return The processed code.
    #
    ##

    if not code:
        return None

    return bleach.clean(
        code,
        tags = validTags,
        strip = True
    )


####
#
# fiction-dl
# Copyright (C) (2020 - 2021) Benedykt Synakiewicz <dreamcobbler@outlook.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####

#
#
#
# Imports.
#
#
#

# Application.

from fiction_dl.Concepts.Chapter import Chapter
from fiction_dl.Concepts.Extractor import Extractor
import fiction_dl.Configuration as Configuration

# Standard packages.

import logging
from typing import List, Optional

# Non-standard packages.

from bs4 import BeautifulSoup
from dreamy_utilities.Web import GetHostname

#
#
#
# Classes.
#
#
#

##
#
# An extractor dedicated for www.asstr.org/~Kristen/.
#
##

class ExtractorAsstrKristen(Extractor):

    def __init__(self) -> None:

        ##
        #
        # The constructor.
        #
        ##

        super().__init__()

        self._downloadStorySoupWhenScanning = False
        self._downloadChapterSoupWhenExtracting = False

        self._storyText = None

    def GetSupportedHostnames(self) -> List[str]:

        ##
        #
        # Returns a list of hostnames supposed to be supported by the extractor.
        #
        # @return A list of supported hostnames.
        #
        ##

        return [
            "asstr.org",
        ]

    def _InternallyScanStory(
        self,
        URL: str,
        soup: Optional[BeautifulSoup]
    ) -> bool:

        ##
        #
        # Scans the story: generates the list of chapter URLs and retrieves the
        # metadata.
        #
        # @param URL  The URL of the story.
        # @param soup The tag soup.
        #
        # @return **False** when the scan fails, **True** when it doesn't fail.
        #
        ##

        # Download the story.

        self._storyText = self._webSession.Get(URL, textEncoding = "ascii")
        if not self._storyText:
            logging.error("Failed to download story page when scanning.")
            return False

        # Locate and read the metadata.

        title = None
        author = None
        summary = None

        storyLines = self._storyText.split("\n")
        storyLines = [x for x in storyLines if x.strip()]

        firstSeparatorLineIndex = -1
        secondSeparatorLineIndex = -1

        for index, line in enumerate(storyLines):

            if (-1 != firstSeparatorLineIndex) and (-1 != secondSeparatorLineIndex):
                break

            elif not line.strip().startswith("***"):
                continue

            elif -1 == firstSeparatorLineIndex:
                firstSeparatorLineIndex = index

            elif -1 == secondSeparatorLineIndex:
                secondSeparatorLineIndex = index

        if (-1 == firstSeparatorLineIndex) or (-1 == secondSeparatorLineIndex):
            logging.error("Failed to locate the metadata.")
            return False

        author = storyLines[firstSeparatorLineIndex - 1][3:]
        if (separatorPosition := author.find("(")) and (-1 != separatorPosition):
            author = author[:separatorPosition].strip()
        if not author:
            logging.error("Failed to read story author.")
            return False

        title = storyLines[firstSeparatorLineIndex - 2]
        if not title:
            logging.error("Failed to read story title.")
            return False

        summary = " ".join(storyLines[firstSeparatorLineIndex + 1:secondSeparatorLineIndex]).strip()

        # Set the metadata.

        self.Story.Metadata.Title = title
        self.Story.Metadata.Author = author

        self.Story.Metadata.DatePublished = "?"
        self.Story.Metadata.DateUpdated = "?"

        self.Story.Metadata.ChapterCount = 1
        self.Story.Metadata.WordCount = 0

        self.Story.Metadata.Summary = summary

        # Set chapter URLs.

        self._chapterURLs = [URL]

        # Return.

        return True

    def _InternallyExtractChapter(
        self,
        URL: str,
        soup: Optional[BeautifulSoup]
    ) -> Optional[Chapter]:

        ##
        #
        # Extracts specific chapter.
        #
        # @param URL  The URL of the page containing the chapter.
        # @param soup The tag soup of the page containing the chapter.
        #
        # @return **True** if the chapter is extracted correctly, **False** otherwise.
        #
        ##

        # Define usual story endings.

        USUAL_ENDINGS = [
            "~~~",
            "~ ~ ~ ",
            "the end",
            "end",
        ]

        # Locate and cut the end.

        text = self._storyText.splitlines()

        separatorLineIndices = []
        endLineIndices = []

        for index, line in enumerate(text):

            strippedLine = line.strip()

            if strippedLine.startswith("***") or strippedLine.startswith("------"):
                separatorLineIndices.append(index)

            lowercaseLine = strippedLine.lower()

            for ending in USUAL_ENDINGS:
                if lowercaseLine.startswith(ending):
                    endLineIndices.append(index)
                    continue

        firstLineIndex = separatorLineIndices[1] if separatorLineIndices else -1
        lastLineIndex = endLineIndices[-1] if endLineIndices else -1

        if -1 == firstLineIndex:
            logging.error("Invalid story content format.")
            return None

        if -1 == lastLineIndex:
            text = text[firstLineIndex + 1:]
        else:
            text = text[firstLineIndex + 1:lastLineIndex]

        # Format the content.

        chapterCode = ""
        currentParagraphCode = ""

        for line in text:

            if not line:

                chapterCode += f"<p>{currentParagraphCode}</p>"
                currentParagraphCode = ""

            else:

                currentParagraphCode += f" {line.strip()}"

        # Return.

        return Chapter(content = chapterCode)



####
#
# fiction-dl
# Copyright (C) (2020 - 2021) Benedykt Synakiewicz <dreamcobbler@outlook.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####

#
#
#
# Imports.
#
#
#

# Application.

from fiction_dl.Concepts.Formatter import Formatter
from fiction_dl.Concepts.Story import Story

# Standard packages.

from pathlib import Path
from subprocess import call, DEVNULL

# Non-standard packages.

from dreamy_utilities.Text import Stringify

#
#
#
# Classes.
#
#
#

##
#
# The PDF formatter.
#
##

class FormatterPDF(Formatter):

    def __init__(self, embedImages: bool = True) -> None:

        ##
        #
        # The constructor.
        #
        # @param embedImages Embed images in the output file.
        #
        ##

        super().__init__(embedImages)

    def FormatAndSave(self, story: Story, filePath: Path) -> bool:

        ##
        #
        # Formats the story and saves it to the output file. This function is currently **not
        # implemented**.
        #
        # @param story    The story to be formatted and saved.
        # @param filePath The path to the output file.
        #
        # @return **True** if the output file was generated and saved without problems, **False**
        #         otherwise.
        #
        ##

        raise NotImplementedError()

    def ConvertFromODT(
        self,
        sourceFilePath: Path,
        outputDirectoryPath: Path,
        converterFilePath: Path
    ) -> bool:

        ##
        #
        # Converts an ODT file to a PDF file. The output file may exist: it will be overwritten if
        # it does.
        #
        # @param sourceFilePath      Path to the ODT file.
        # @param outputDirectoryPath Path to the output directory. The output file will be created
        #                            inside it; its (base)name will be the same as the name of the
        #                            source file. The directory **has** to exist beforehand, this
        #                            method does *not* create it.
        # @param converterFilePath   Path to the LibreOffice executable (soffice.exe/soffice).
        #
        # @return **True** if the conversion has been performed successfully, **False** otherwise.
        #
        ##

        if not sourceFilePath.is_file():
            return False

        elif not outputDirectoryPath.is_dir():
            return False

        elif not converterFilePath.is_file():
            return False

        call(
            [
                Stringify(converterFilePath),
                "--headless",
                "--convert-to",
                "pdf",
                Stringify(sourceFilePath),
                "--outdir",
                Stringify(outputDirectoryPath),
            ],
            stdout = DEVNULL
        )

        return True


####
#
# fiction-dl
# Copyright (C) (2020 - 2021) Benedykt Synakiewicz <dreamcobbler@outlook.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####

#
#
#
# Imports.
#
#
#

# Application.

from fiction_dl.Concepts.Formatter import Formatter
from fiction_dl.Concepts.Story import Story
from fiction_dl.Concepts.StoryPackage import StoryPackage
from fiction_dl.Utilities.Filesystem import GetPackageDirectory

# Standard packages.

from base64 import b64encode
from pathlib import Path
from typing import List, Union

# Non-standard packages.

from bs4 import BeautifulSoup
from dreamy_utilities.Filesystem import ReadTextFile

#
#
#
# Classes.
#
#
#

##
#
# The HTML formatter.
#
##

class FormatterHTML(Formatter):

    def __init__(self, embedImages: bool = True) -> None:

        ##
        #
        # The constructor.
        #
        # @param embedImages Embed images in the output file.
        #
        ##

        super().__init__(embedImages)

    def FormatAndSave(self, story: Union[Story, StoryPackage], filePath: Path) -> bool:

        ##
        #
        # Formats the story (or the story package) and saves it to the output file.
        #
        # @param story    The story/story package to be formatted and saved.
        # @param filePath The path to the output file.
        #
        # @return **True** if the output file was generated and saved without problems, **False**
        #         otherwise.
        #
        ##

        # Choose the appropriate template file.

        templateFileName =                          \
            "FormatterHTML (Package).html"          \
            if isinstance(story, StoryPackage) else \
            "FormatterHTML.html"

        # Load the template and fill it with the story.

        templateFilePath = GetPackageDirectory() / f"Templates/{templateFileName}"
        content = story.FillTemplate(ReadTextFile(templateFilePath), escapeHTMLEntities = True)

        def ChapterTitler(index: int, chapterTitle: str) -> str:
            return f"Chapter {index}" + (f": {chapterTitle}" if chapterTitle else "")

        def PackagePrefixer(index: int, chapterTitle: str, storyTitle: str) -> str:
            return f"<h2>{storyTitle} — {ChapterTitler(index, chapterTitle)}</h2>"

        def StoryPrefixer(index: int, chapterTitle: str, storyTitle: str) -> str:
            return f"<h2>{ChapterTitler(index, chapterTitle)}</h2>"

        prefixer = PackagePrefixer if isinstance(story, StoryPackage) else StoryPrefixer
        content = content.replace(
            "@@@Content@@@",
            story.Join(prefixer)
        )

        # Replace images.

        if self._embedImages:

            soup = BeautifulSoup(content, features = "html.parser")

            for index, tag in enumerate(soup.find_all("img")):

                if (0 == index) or (index > len(story.Images)):
                    continue

                if (image := story.Images[index - 1]):
                    tag["src"] = "data:image/jpeg;base64," + b64encode(image.Data).decode()
                    tag["alt"] = "There is an image here."
                else:
                    tag["alt"] = "There ought to be an image here."

            content = str(soup)

        else:

            content = content.replace("<img/>", "")

        # Save the template to file.

        try:

            with open(filePath, "w", encoding = "utf-8") as outputFile:
                outputFile.write(content)

            return True

        except OSError:

            return False


####
#
# fiction-dl
# Copyright (C) (2020 - 2021) Benedykt Synakiewicz <dreamcobbler@outlook.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####

#
#
#
# Imports.
#
#
#

# Application.

from fiction_dl.Concepts.Chapter import Chapter
from fiction_dl.Concepts.Extractor import Extractor
from fiction_dl.Concepts.Story import Story

# Standard packages.

import logging
import re
from typing import List, Optional, Tuple

# Non-standard packages.

from bs4 import BeautifulSoup, Comment, NavigableString
from dreamy_utilities.Text import FindFirstMatch, Stringify
from dreamy_utilities.Web import GetHostname

#
#
#
# Classes.
#
#
#

##
#
# An extractor dedicated for WhoFic.com.
#
##

class ExtractorWhoFic(Extractor):

    def __init__(self) -> None:

        ##
        #
        # The constructor.
        #
        ##

        super().__init__()

    def GetSupportedHostnames(self) -> List[str]:

        ##
        #
        # Returns a list of hostnames supposed to be supported by the extractor.
        #
        # @return A list of supported hostnames.
        #
        ##

        return [
            "whofic.com",
        ]

    def ScanChannel(self, URL: str) -> Optional[List[str]]:

        ##
        #
        # Scans the channel: generates the list of story URLs.
        #
        # @return **None** when the scan fails, a list of story URLs when it doesn't fail.
        #
        ##

        if (not URL) or (GetHostname(URL) not in self.GetSupportedHostnames()):
            return None

        if ("seriesid" not in URL) and ("viewuser.php" not in URL):
            return None

        soup = self._webSession.GetSoup(URL)
        if not soup:
            logging.error(f"Failed to download page: \"{URL}\".")
            return None

        return self._GetStoryURLsOnPage(soup)

    def _InternallyScanStory(
        self,
        URL: str,
        soup: Optional[BeautifulSoup]
    ) -> bool:

        ##
        #
        # Scans the story: generates the list of chapter URLs and retrieves the
        # metadata.
        #
        # @param URL  The URL of the story.
        # @param soup The tag soup.
        #
        # @return **False** when the scan fails, **True** when it doesn't fail.
        #
        ##

        # Locate relevant page elements.

        titleElement = soup.select_one("div#storyHeader > h1")
        if not titleElement:
            logging.error("Title element not found.")
            return False

        authorElement = soup.select_one("div#storyHeader > p.mb-0 > a")
        if not authorElement:
            logging.error("Author element not found.")
            return False

        # Extract chapter URLs.

        for element in soup.select("div.container > div.row > div.box > p > b > a"):

            if not element.has_attr("href"):
                continue

            self._chapterURLs.append(self._BASE_URL + element["href"])

        if not self._chapterURLs:
            self._chapterURLs.append(URL)

        # Find additional metadata.

        datePublished, dateUpdated, summary = self._FindAdditionalMetadata(
            ExtractorWhoFic._BASE_URL + authorElement["href"],
            self._GetStoryID(URL)
        )

        # Set the metadata.

        self.Story.Metadata.Title = titleElement.get_text().strip()
        self.Story.Metadata.Author = authorElement.get_text().strip()

        self.Story.Metadata.DatePublished = datePublished
        self.Story.Metadata.DateUpdated = dateUpdated

        self.Story.Metadata.ChapterCount = len(self._chapterURLs)
        self.Story.Metadata.WordCount = 0

        self.Story.Metadata.Summary = summary

        # Return.

        return True

    def _InternallyExtractChapter(
        self,
        URL: str,
        soup: Optional[BeautifulSoup]
    ) -> Optional[Chapter]:

        ##
        #
        # Extracts specific chapter.
        #
        # @param URL  The URL of the page containing the chapter.
        # @param soup The tag soup of the page containing the chapter.
        #
        # @return **True** if the chapter is extracted correctly, **False** otherwise.
        #
        ##

        # Locate relevant page elements.

        contentElement = soup.select_one("div.container > div.row > div.box")
        if not contentElement:
            logging.error("Content element not found.")
            return False

        if (element := contentElement.select_one("div#storyHeader")):
            element.decompose()

        if (element := contentElement.select_one("div#authorNotes")):
            element.decompose()

        for element in contentElement.select("form"):
            element.decompose()

        # Return.

        return Chapter(
            content = Stringify(contentElement.encode_contents())
        )

    @staticmethod
    def _GetStoryID(URL: str) -> Optional[str]:

        ##
        #
        # Retrieves story ID from story URL.
        #
        # @param URL The URL of the story.
        #
        # @return The ID of the story. Optionally **None**.
        #
        ##

        if not URL:
            return None

        return FindFirstMatch(URL, "sid=(\d+)")

    def _FindAdditionalMetadata(self, authorPageURL: str, storyID: str) -> Tuple[str, str, str]:

        ##
        #
        # Reads story's publication and update dates from the author's page, as well as its summary.
        #
        # @param authorPageURL The URL of the author's profile page.
        # @param storyID       The ID of the story.
        #
        # @return A tuple consisting of date published, date updated and summary.
        #
        ##

        if (not authorPageURL) or (not storyID):
            return None

        # Prepare the default output.

        datePublished = "?"
        dateUpdated = "?"
        summary = "No summary."

        # Download the author's page soup.

        soup = self._webSession.GetSoup(authorPageURL)
        if not soup:
            return [datePublished, dateUpdated, summary]

        # Find this specific story's description.

        storyElement = None

        for storyBlockElement in soup.select("div.storyBlock"):

            anchorElement = storyBlockElement.select_one("p > strong > a")
            if (not anchorElement) or (not anchorElement.has_attr("href")):
                continue

            ID = ExtractorWhoFic._GetStoryID(anchorElement["href"])
            if not ID:
                continue

            if ID == storyID:
                storyElement = storyBlockElement
                break

        if not storyElement:
            return [datePublished, dateUpdated, summary]

        # Process the dates.

        for listItemElement in storyElement.select("ul.list-inline > li"):

            titleElement = listItemElement.select_one("b")
            if not titleElement:
                continue

            titleElementText = titleElement.get_text().strip()
            titleElement.decompose()

            if "Published:" == titleElementText:
                datePublished = listItemElement.get_text().strip()
            elif "Updated:" == titleElementText:
                dateUpdated = listItemElement.get_text().strip()

        # Process the summary.

        potentialSummaryStrings = []

        if (summaryParentElement := storyElement.select_one("p")):

            for child in summaryParentElement.children:

                if isinstance(child, NavigableString) and not isinstance(child, Comment):
                    potentialSummaryStrings.append(str(child).strip())

        if potentialSummaryStrings:
            summary = potentialSummaryStrings[-1]

        # Post-process found metadata.

        if "?" != datePublished:
            datePublished = ExtractorWhoFic._ReformatDate(datePublished)

        if "?" != dateUpdated:
            dateUpdated = ExtractorWhoFic._ReformatDate(dateUpdated)

        # Return.

        return [datePublished, dateUpdated, summary]

    @staticmethod
    def _GetStoryURLsOnPage(soup: BeautifulSoup) -> List[str]:

        ##
        #
        # Reads all the story links present on a specific page.
        #
        # @param soup The input tag soup.
        #
        # @return A list of (absolute) story URLs.
        #
        ##

        URLs = []

        for element in soup.select("div.box > div.storyBlock > p > strong > a"):

            if not element.has_attr("href"):
                continue

            storyIDMatch = re.search(
                "sid=(\d+)",
                element["href"]
            )

            if not storyIDMatch:
                continue

            storyID = storyIDMatch.group(1)
            storyURL = f"{ExtractorWhoFic._BASE_URL}viewstory.php?sid={storyID}"

            URLs.append(storyURL)

        return URLs

    @staticmethod
    def _ReformatDate(date: str) -> Optional[str]:

        ##
        #
        # Reformats date from YYYY.MM.DD to YYYY-MM-DD.
        #
        # @param date The input date.
        #
        # @return The input date, reformatted.
        #
        ##

        if not date:
            return None

        return date.replace(".", "-")

    _BASE_URL = "https://www.whofic.com/"


####
#
# fiction-dl
# Copyright (C) (2020 - 2021) Benedykt Synakiewicz <dreamcobbler@outlook.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####

#
#
#
# Imports.
#
#
#

# Application.

from fiction_dl.Concepts.Formatter import Formatter
from fiction_dl.Concepts.Story import Story
from fiction_dl.Concepts.StoryPackage import StoryPackage
from fiction_dl.Utilities.Filesystem import GetPackageDirectory
from fiction_dl.Utilities.HTML import ReformatHTMLToXHTML

# Standard packages.

from pathlib import Path
from typing import List, Optional, Union

# Non-standard packages.

from bs4 import BeautifulSoup
from dreamy_utilities.Filesystem import ReadTextFile
from ebooklib import epub

#
#
#
# Classes.
#
#
#

##
#
# The EPUB formatter.
#
##

class FormatterEPUB(Formatter):

    def __init__(
        self,
        embedImages: bool = True,
        coverImageData: Optional[bytes] = None
    ) -> None:

        ##
        #
        # The constructor.
        #
        # @param embedImages Embed images in the output file.
        #
        ##

        super().__init__(embedImages)

        self.CoverImageData = coverImageData

    def FormatAndSave(self, story: Union[Story, StoryPackage], filePath: Path) -> bool:

        ##
        #
        # Formats the story (or the story package) and saves it to the output file.
        #
        # @param story    The story/story package to be formatted and saved.
        # @param filePath The path to the output file.
        #
        # @return **True** if the output file was generated and saved without problems, **False**
        #         otherwise.
        #
        ##

        # Create the e-book.

        book = epub.EpubBook()

        metadata = story.Metadata.GetPrettified()

        book.set_identifier(metadata.Title)
        book.set_title(metadata.Title)
        book.set_language("en")
        book.add_metadata("DC", "description", metadata.Summary)
        book.add_author(metadata.Author)

        # Set up the cover.

        if self.CoverImageData:
            book.set_cover(self._CoverImageName, self.CoverImageData)

        # Embed images.

        if self._embedImages:

            for index, image in enumerate(story.Images):

                if not image:
                    continue

                self._EmbedImage(book, f"{index}.jpeg", image.Data)

        # Prepare the spine.

        book.spine = ["cover", "nav"]

        # Add the metadata chapter.

        templateFileName =                          \
            "Metadata (Package).html"               \
            if isinstance(story, StoryPackage) else \
            "Metadata.html"

        metadataFilePath = GetPackageDirectory() / f"Templates/FormatterEPUB/{templateFileName}"
        metadataTemplate = ReadTextFile(metadataFilePath)

        metadataChapter = epub.EpubHtml(
            file_name = "Metadata.xhtml",
            content = story.FillTemplate(metadataTemplate, escapeHTMLEntities = True),
            title = "Metadata",
            lang = "en"
        )

        book.add_item(metadataChapter)
        book.spine.append(metadataChapter)

        # Add the stylesheet.

        stylesheetFilePath = GetPackageDirectory() / "Templates/FormatterEPUB/Stylesheet.css"

        stylesheet = epub.EpubItem(
            uid = "style_default",
            file_name = "style/default.css",
            media_type = "text/css",
            content = ReadTextFile(stylesheetFilePath)
        )

        book.add_item(stylesheet)

        # Add chapters and create book spine.

        imageIndex = 0
        stories = story.Stories if isinstance(story, StoryPackage) else [story]

        for storyIndex, specificStory in enumerate(stories, start = 1):

            prettifiedStoryTitle = specificStory.Metadata.GetPrettified().Title

            for index, chapter in enumerate(specificStory.Chapters, start = 1):

                def PackageChapterTitler(index: int, chapterTitle: str, storyTitle: str) -> str:
                    return f"{storyTitle} — Chapter {index}" + (f": {chapterTitle}" if chapterTitle else "")

                def StoryChapterTitler(index: int, chapterTitle: str, storyTitle: str) -> str:
                    return f"Chapter {index}" + (f": {chapterTitle}" if chapterTitle else "")

                def PackagePrefixer(index: int, chapterTitle: str, storyTitle: str) -> str:
                    return f"<h2>{PackageChapterTitler(index, chapterTitle, storyTitle)}</h2>"

                def StoryPrefixer(index: int, chapterTitle: str, storyTitle: str) -> str:
                    return f"<h2>{StoryChapterTitler(index, chapterTitle, storyTitle)}</h2>"

                titler = PackageChapterTitler if isinstance(story, StoryPackage) else StoryChapterTitler
                prefixer = PackagePrefixer if isinstance(story, StoryPackage) else StoryPrefixer

                # Create the chapter.

                bookChapter = epub.EpubHtml(
                    file_name = f"Story {storyIndex} - Chapter {index}.xhtml",
                    title = titler(index, chapter.Title, prettifiedStoryTitle),
                    lang = "en"
                )

                # Prepare chapter content.

                content = prefixer(index, chapter.Title, prettifiedStoryTitle) + chapter.Content
                content = ReformatHTMLToXHTML(content)

                if self.CoverImageData and (len(specificStory.Chapters) == index):
                    content += f'<img src = "{self._CoverImageName}"/>'

                # Replace images.

                if self._embedImages:

                    soup = BeautifulSoup(content, features = "html.parser")

                    for index, tag in enumerate(soup.find_all("img")):

                        if index >= len(specificStory.Images):
                            continue

                        if (image := specificStory.Images[index]):
                            tag["src"] = f"{index}.jpeg"
                            tag["alt"] = "There is an image here."
                        else:
                            tag["alt"] = "There ought to be an image here."

                    content = str(soup)

                else:

                    content = content.replace("<img/>", "")

                # Set the chapter.

                bookChapter.set_content(content)

                book.add_item(bookChapter)
                book.spine.append(bookChapter)

        # Create a ToC.

        book.toc = book.spine[1:]

        # Add NCX and Navigation tile.

        book.add_item(epub.EpubNcx())
        book.add_item(epub.EpubNav())

        # Write the file.

        epub.write_epub(filePath, book, {})

        # Return.

        return True

    @staticmethod
    def _EmbedImage(
        book: epub.EpubBook,
        fileName: str,
        data: bytes,
        fileType: str = "image/jpeg"
    ) -> bool:

        ##
        #
        # Embeds an image inside the created ebook.
        #
        # @param book     The ebook.
        # @param fileName Embedded image's file name (that is, the file name of the image file
        #                 *inside* the ebook archive).
        # @param data     Image data: binary and encoded.
        # @param fileType Image data type.
        #
        # @return **True** if the image has been embedded correctly, **False** otherwise.
        #
        ##

        if (not book) or (not fileName) or (not data) or (not fileType):
            return False

        item = epub.EpubImage()
        item.file_name = fileName
        item.media_type = fileType
        item.content = data

        book.add_item(item)

        return True

    _CoverImageName = "Cover.jpeg"


####
#
# fiction-dl
# Copyright (C) (2020 - 2021) Benedykt Synakiewicz <dreamcobbler@outlook.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####

#
#
#
# Imports.
#
#
#

# Application.

from fiction_dl.Concepts.Chapter import Chapter
from fiction_dl.Concepts.Extractor import Extractor
from fiction_dl.Concepts.Story import Story
import fiction_dl.Configuration as Configuration

# Standard packages.

from itertools import groupby
from typing import List, Optional

# Non-standard packages.

from dreamy_utilities.Filesystem import ReadTextFile

#
#
#
# Classes.
#
#
#

class ExtractorTextFile(Extractor):

    def __init__(self) -> None:

        ##
        #
        # The constructor.
        #
        ##

        super().__init__()

        self._filePath = None
        self._fileContent = None
        self._chapters = None

    def RequiresBreaksBetweenRequests(self) -> bool:

        ##
        #
        # Does the extractor require the application to sleep between subsequent reqests?
        #
        # @return **True** if it does, **False** otherwise.
        #
        ##

        return False

    def Initialize(self, filePath: str) -> bool:

        ##
        #
        # Initializes the extractor with given URL.
        #
        # @param filePath The URL in question.
        #
        # @return **True** if the extractor can handle the URL, **False** otherwise.
        #
        ##

        if not ReadTextFile(filePath, lines = True):
            return False

        self._filePath = filePath
        self.Story = Story()

        return True

    def ScanChannel(self, URL: str) -> Optional[List[str]]:

        ##
        #
        # Scans the channel: generates the list of story URLs.
        #
        # @return **None** when the scan fails, a list of story URLs when it doesn't fail.
        #
        ##

        lines = ReadTextFile(self._filePath, lines = True)
        if (not lines) or lines[0].startswith(Configuration.TextSourceFileMagicText):
            return None

        return self._ReadURLsFromLines(lines)

    def ScanStory(self) -> bool:

        ##
        #
        # Scans the story: generates the list of chapter URLs and retrieves the
        # metadata.
        #
        # @return **False** when the scan fails, **True** when it doesn't fail.
        #
        ##

        if not self._filePath:
            return False

        # Read the file.

        self._fileContent = ReadTextFile(self._filePath, lines = True)
        if (not self._fileContent) or len(self._fileContent) < 6:
            return False

        # Calculate chapter count.

        chapterCount = 1

        for line in self._fileContent:
            if line.startswith(Configuration.TextSourceFileChapterBreak):
                chapterCount += 1

        # Set the metadata.

        self.Story.Metadata.URL = self._fileContent[1].strip()
        self.Story.Metadata.Title = self._fileContent[2].strip()
        self.Story.Metadata.Author = self._fileContent[3].strip()
        self.Story.Metadata.Summary = self._fileContent[4].strip()
        self.Story.Metadata.ChapterCount = chapterCount
        self.Story.Metadata.WordCount = 0

        # Format chapter content.

        chapterGroups = groupby(
            self._fileContent[5:],
            lambda x: x.startswith(Configuration.TextSourceFileChapterBreak)
        )

        self._chapters = [list(group) for x, group in chapterGroups if not x]

        # Return.

        return True

    def ExtractChapter(self, index: int) -> Optional[Chapter]:

        ##
        #
        # Extracts specific chapter.
        #
        # @param index The index of the chapter to be extracted.
        #
        # @return **True** if the chapter is extracted correctly, **False** otherwise.
        #
        ##

        if (not self._chapters) or (index < 1):
            return None

        return Chapter(content = "".join(self._chapters[index - 1]))

    @staticmethod
    def _ReadURLsFromLines(lines: List[str]) -> List[str]:

        ##
        #
        # Creates a list of input URLs from lines of a text file.
        #
        # @param lines List of strings being read lines of a text file.
        #
        # @return List of strings: URLs of the stories to be downloaded.
        #
        ##

        if not lines:
            return lines

        URLs = [x.strip() for x in lines]
        URLs = [x for x in URLs if len(x)] # Ignore empty lines.
        URLs = [x for x in URLs if not x.startswith("#")] # Ignore comments.

        return URLs


####
#
# fiction-dl
# Copyright (C) (2020 - 2021) Benedykt Synakiewicz <dreamcobbler@outlook.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####

#
#
#
# Imports.
#
#
#

# Application.

from fiction_dl.Concepts.Chapter import Chapter
from fiction_dl.Concepts.Extractor import Extractor
from fiction_dl.Concepts.Story import Story
from fiction_dl.Utilities.Text import GetTitleProper

# Standard packages.

import logging
import re
import requests
from typing import List, Optional

# Non-standard packages.

from bs4 import BeautifulSoup
from dreamy_utilities.Text import (
    GetLevenshteinDistance,
    GetLongestLeadingSubstring,
    PrettifyTitle,
    SeparateSubtitle,
    Stringify
)
from dreamy_utilities.Web import GetHostname

#
#
#
# Classes.
#
#
#

##
#
# An extractor dedicated for NajlepszaErotyka.com.pl.
#
##

class ExtractorNajlepszaErotyka(Extractor):

    def __init__(self) -> None:

        ##
        #
        # The constructor.
        #
        ##

        super().__init__()

        self._chapterParserName = "html5lib"

    def GetSupportedHostnames(self) -> List[str]:

        ##
        #
        # Returns a list of hostnames supposed to be supported by the extractor.
        #
        # @return A list of supported hostnames.
        #
        ##

        return [
            "najlepszaerotyka.com.pl",
        ]

    def _InternallyScanStory(
        self,
        URL: str,
        soup: Optional[BeautifulSoup]
    ) -> bool:

        ##
        #
        # Scans the story: generates the list of chapter URLs and retrieves the
        # metadata.
        #
        # @param URL  The URL of the story.
        # @param soup The tag soup.
        #
        # @return **False** when the scan fails, **True** when it doesn't fail.
        #
        ##

        # Locate relevant elements.

        titleElement = soup.select_one(".container-fluid .row h2")
        if not titleElement:
            logging.error("Title element not found.")
            return False

        titleSpanElement = titleElement.find("span")
        titleSpanElement.replaceWith("")

        metadataElements = soup.select(".container-fluid .row div ul li")
        if not metadataElements:
            logging.error("Author element not found.")
            return False

        authorElement = metadataElements[-1].find("a")
        dateElement = metadataElements[-2]

        # Process metadata.

        authorNameMatch = re.search("/author/(.+)/", authorElement["href"])
        if not authorNameMatch:
            logging.error("Failed to retrieve author's name.")
            return False

        title = self._CleanStoryTitle(titleElement.get_text().strip())
        titleProper = GetTitleProper(title)

        # Find all other chapters of this story.

        self._chapterURLs = []
        chapterTitles = []

        for story in self._FindAllStoriesByAuthor(authorNameMatch.group(1)):

            currentTitleProper = GetTitleProper(story[0])

            distance = GetLevenshteinDistance(currentTitleProper, titleProper)
            if distance > 5:
                continue

            self._chapterURLs.append([story[1], story[2]])
            chapterTitles.append(currentTitleProper)

        datePublished = dateElement.get_text().strip()
        dateUpdated = dateElement.get_text().strip()

        if not self._chapterURLs:

            self._chapterURLs = [self.Story.Metadata.URL]

        else:

            datePublished = self._chapterURLs[0][1]
            dateUpdated = self._chapterURLs[-1][1]

            self._chapterURLs = [x[0] for x in self._chapterURLs]

        # Set the metadata.

        self.Story.Metadata.Title = GetLongestLeadingSubstring(chapterTitles).strip()
        self.Story.Metadata.Author = authorElement.get_text().strip()

        self.Story.Metadata.DatePublished = datePublished
        self.Story.Metadata.DateUpdated = dateUpdated

        self.Story.Metadata.ChapterCount = len(self._chapterURLs)
        self.Story.Metadata.WordCount = 0

        self.Story.Metadata.Summary = "No summary."

        # Return.

        return True

    def _InternallyExtractChapter(
        self,
        URL: str,
        soup: Optional[BeautifulSoup]
    ) -> Optional[Chapter]:

        ##
        #
        # Extracts specific chapter.
        #
        # @param URL  The URL of the page containing the chapter.
        # @param soup The tag soup of the page containing the chapter.
        #
        # @return **True** if the chapter is extracted correctly, **False** otherwise.
        #
        ##

        titleElement = soup.select_one(".container-fluid .row h2")
        if not titleElement:
            logging.error("Title element not found.")
            return False

        if (unwantedElement := titleElement.find("span")):
            unwantedElement.replaceWith("")

        contentElement = soup.select_one(".container .row .entry-content")
        if not contentElement:
            logging.error("Content element not found.")
            return None

        if (unwantedElement := contentElement.select_one("span.rt-reading-time")):
            unwantedElement.replaceWith("")

        if (unwantedElement := contentElement.select_one("div.wpcm-subscribe")):
            unwantedElement.replaceWith("")

        if (unwantedElement := contentElement.select_one("rating-form")):
            unwantedElement.replaceWith("")

        return Chapter(
            title = SeparateSubtitle(self._CleanStoryTitle(titleElement.get_text().strip())),
            content = Stringify(contentElement.encode_contents())
        )

    def _FindAllStoriesByAuthor(self, authorName: str):

        # Download author's page.

        authorsPageURL = f"https://najlepszaerotyka.com.pl/author/{authorName}/"

        soup = self._webSession.GetSoup(authorsPageURL)
        if not soup:
            logging.error("Failed to download page: \"{authorsPageURL\".")
            return None

        # Get the number of subpages.

        pageCount = 1

        paginationElements = soup.select("a.page-numbers")
        if paginationElements:
            pageCount = len(paginationElements)

        # Go over all the pages and collect story links and titles.

        stories = []

        for index in range(1, pageCount + 1):

            pageURL = f"https://najlepszaerotyka.com.pl/author/{authorName}/page/{index}"

            soup = self._webSession.GetSoup(pageURL)
            if not soup:
                logging.error("Failed to download page: \"{pageURL\".")
                return None

            postElements = soup.select("div.post > div.blog-details")

            for element in postElements:

                linkElement = element.select_one("h2 > a")
                if (not linkElement) or (not linkElement.has_attr("href")):
                    continue

                title = ExtractorNajlepszaErotyka._CleanStoryTitle(linkElement.get_text().strip())
                link = linkElement["href"]

                dateElement = element.select_one("li.published")
                date = dateElement.get_text().strip()

                stories.append([title, link, date])

        # Return.

        stories.reverse()
        return stories

    @staticmethod
    def _CleanStoryTitle(title: str) -> str:

        ##
        #
        # Cleans story title (removes recurring elements).
        #
        # @param title The input title.
        #
        # @return The cleaned title.
        #
        ##

        titleMatch = re.search("(.+) \(.+\)", title)
        if not titleMatch:
            return title

        return titleMatch.group(1)


####
#
# fiction-dl
# Copyright (C) (2020 - 2021) Benedykt Synakiewicz <dreamcobbler@outlook.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####

#
#
#
# Imports.
#
#
#

# Application.

from fiction_dl.Concepts.Chapter import Chapter
from fiction_dl.Concepts.Extractor import Extractor

# Standard packages.

import logging
import re
from typing import List, Optional

# Non-standard packages.

from bs4 import BeautifulSoup
from dreamy_utilities.Filesystem import WriteTextFile
from dreamy_utilities.Interface import Interface
from dreamy_utilities.Text import GetDateFromTimestamp, Stringify
from dreamy_utilities.Web import GetSiteURL

#
#
#
# The class definition.
#
#
#

class ExtractorXenForo(Extractor):

    def __init__(self, forumURL: str) -> None:

        ##
        #
        # The constructor.
        #
        ##

        super().__init__()

        self._baseURL = GetSiteURL(forumURL)
        self._forumURL = forumURL

    def SupportsAuthentication(self) -> bool:

        ##
        #
        # Checks whether the extractor supports user authentication.
        #
        # @return **True** if the site *does* support authentication, **False** otherwise.
        #
        ##

        return True

    def Authenticate(self, interface: Interface) -> bool:

        ##
        #
        # Logs the user in, interactively.
        #
        # @param interface The user interface to be used.
        #
        # @return **True** if the user has been authenticated correctly, **False** otherwise.
        #
        ##

        # Format the log-in URL.

        LOGIN_URL = self._forumURL + "login/login"

        # Download the log-in page.

        soup = self._webSession.GetSoup(LOGIN_URL)
        if not soup:
            logging.error("Failed to download the log-in page.")
            return self.AuthenticationResult.FAILURE

        formElement = soup.select_one("div.blocks > form.block")
        if not formElement:
            logging.error("Log-in form element not found.")
            return self.AuthenticationResult.FAILURE

        authenticityTokenElement = formElement.find("input", {"name": "_xfToken"})
        if not authenticityTokenElement:
            logging.error("\"_xfToken\" input field not found.")
            return self.AuthenticationResult.FAILURE
        elif not authenticityTokenElement.has_attr("value"):
            logging.error("\"_xfToken\" input field doesn't have a value.")
            return self.AuthenticationResult.FAILURE

        authenticityToken = authenticityTokenElement["value"].strip()

        # Read the username and the password.

        userName = ""
        password = ""

        if ExtractorXenForo._userName is None:

            interface.GrabUserAttention()

            userName = interface.ReadString("Your username")
            if userName:
                password = interface.ReadPassword("Your password")

        else:

            userName = ExtractorXenForo._userName
            password = ExtractorXenForo._userPassword

        if not userName:
            userName = ""

        # Remember the username and the password.

        ExtractorXenForo._userName = userName
        ExtractorXenForo._userPassword = password

        # Decide whether to log-in.

        if (not userName) or (not password):
            return self.AuthenticationResult.ABANDONED

        # Attempt to log-in.

        data = {
            "login": userName,
            "password": password,
            "register": 0,
            "remember": "1",
            "cookie_check": "1",
            "_xfToken": authenticityToken,
        }

        response = self._webSession.Post(
            LOGIN_URL,
            data
        )

        # Verify the response and return.

        if not response:
            return self.AuthenticationResult.FAILURE

        return self.AuthenticationResult.SUCCESS

    def ExtractChapter(self, index: int) -> Optional[Chapter]:

        ##
        #
        # Extracts specific chapter.
        #
        # @param index The index of the chapter to be extracted.
        #
        # @return **True** if the chapter is extracted correctly, **False** otherwise.
        #
        ##

        if index > len(self._chapterURLs):
            logging.error(
                f"Trying to extract chapter {index}. "
                f"Only {len(self._chapterURLs)} chapter(s) located. "
                f"The story supposedly has {self.Story.Metadata.ChapterCount} chapter(s)."
            )
            return None

        chapterURL = self._chapterURLs[index - 1]

        soup = self._webSession.GetSoup(chapterURL)
        if not soup:
            logging.error(f'Failed to download page: "{chapterURL}".')
            return None

        if -1 != (postIDLocation := chapterURL.find("#")):

            postID = chapterURL[postIDLocation + 1:]

            postElement = soup.find("article", {"data-content": postID}) or soup.select_one(f"#{postID}")
            if not postElement:
                logging.error("Post element not found.")
                return None

            bodyElement = postElement.select_one("div.bbWrapper") or postElement.select_one("div.messageContent")
            if not bodyElement:
                logging.error("Message body element not found.")
                return None

        else:

            postElement = soup.select_one("li.message")
            if not postElement:
                logging.error("Post element not found.")
                return None

            bodyElement = postElement.select_one("div.messageContent")
            if not bodyElement:
                logging.error("Message body element not found.")
                return None

        return Chapter(content = Stringify(bodyElement.encode_contents()))

    def _InternallyScanStory(
        self,
        URL: str,
        soup: Optional[BeautifulSoup]
    ) -> bool:

        ##
        #
        # Scans the story: generates the list of chapter URLs and retrieves the
        # metadata.
        #
        # @param URL  The URL of the story.
        # @param soup The tag soup.
        #
        # @return **False** when the scan fails, **True** when it doesn't fail.
        #
        ##

        # Generate the threadmarks URL.

        threadmarksURL = self._GetThreadmarksURL(self.Story.Metadata.URL)
        if not threadmarksURL:
            logging.error("Failed to generate threadmarks URL.")
            return False

        # Retrieve story metadata.

        soup = self._webSession.GetSoup(threadmarksURL)
        if not soup:
            logging.error(f'Failed to download page: "{threadmarksURL}".')
            return False

        titleElement = soup.select_one("h1.p-title-value") or soup.select_one(".titleBar > h1")
        if not titleElement:
            logging.error("Title element not found.")
            return False

        titleSpanElements = titleElement.find_all("span")
        for element in titleSpanElements:
            element.decompose()

        authorElements = soup.find_all(attrs = {"data-content-author": True})
        if not authorElements:
            logging.error("Author elements not found.")
            return False

        datePublished = int(authorElements[0]["data-content-date"].strip())
        dateUpdated = int(authorElements[-1]["data-content-date"].strip())

        self.Story.Metadata.Title = titleElement.get_text().strip()
        if self.Story.Metadata.Title.endswith("- Threadmarks"):
            self.Story.Metadata.Title = self.Story.Metadata.Title[:-14]
        elif self.Story.Metadata.Title.startswith("Threadmarks for:"):
            self.Story.Metadata.Title = self.Story.Metadata.Title[17:]

        titleProperMatch = re.search("(.+) \(.+\)", self.Story.Metadata.Title)
        if titleProperMatch:
            self.Story.Metadata.Title = titleProperMatch.group(1)

        self.Story.Metadata.Author = authorElements[0]["data-content-author"].strip()
        self.Story.Metadata.Summary = "No summary."

        self.Story.Metadata.DatePublished = GetDateFromTimestamp(datePublished)
        self.Story.Metadata.DateUpdated = GetDateFromTimestamp(dateUpdated)

        self.Story.Metadata.ChapterCount = len(authorElements)
        self.Story.Metadata.WordCount = 0

        # Retrieve chapter URLs.

        chapterLinkElements = soup.select("div.structItem-title a")
        if not chapterLinkElements:
            chapterLinkElements = soup.select(".threadmarkList > ol > li > a")

        for element in chapterLinkElements:

            chapterRelativeURL = element["href"]
            if not chapterRelativeURL.startswith("/"):
                chapterRelativeURL = "/" + chapterRelativeURL

            chapterURL = self._baseURL + chapterRelativeURL

            self._chapterURLs.append(chapterURL)

        if not self._chapterURLs:
            logging.error("Failed to retrieve chapter URLs.")
            return False

        # Return.

        return True

    def _GetThreadmarksURL(self, URL: str) -> Optional[str]:

        ##
        #
        # Retrieves a URL leading to the threadmarks page.
        #
        # @param URL The URL of the story.
        #
        # @return URL leading to the threadmarks page.
        #
        ##

        if not URL:
            return None

        elif URL.endswith("threadmarks"):
            return URL

        threadTitleMatch = re.search("/threads/(.*)/", URL)
        if not threadTitleMatch:
            return None

        threadTitle = threadTitleMatch.group(1)

        return f"{self._forumURL}threads/{threadTitle}/threadmarks"

    _userName = None
    _userPassword = None


####
#
# fiction-dl
# Copyright (C) (2020 - 2021) Benedykt Synakiewicz <dreamcobbler@outlook.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
####

#
#
#
# Imports.
#
#
#

# Application.

from fiction_dl.Utilities.Extractors import CreateExtractor
import fiction_dl.Configuration as Configuration

# Standard packages.

import random
from typing import List

# Non-standard packages.

from dreamy_utilities.Containers import RemoveDuplicates

#
#
#
# Classes.
#
#
#

##
#
# Represents the input arguments (URLs or file paths).
#
##

class InputData:

    def __init__(self, initialArgument: str) -> None:

        ##
        #
        # The constructor.
        #
        # @param initialArgument The original argument provided by the user.
        #
        ##

        self._expandedArguments = [initialArgument]

    def Access(self) -> List[str]:

        ##
        #
        # Returns the list of story URLs found in the input data.
        #
        # @return A list of story URLs.
        #
        ##

        return self._expandedArguments

    def Expand(self) -> int:

        ##
        #
        # Expands every input argument that can be expanded.
        #
        # @return The number of arguments that have been expanded.
        #
        ##

        if not self._expandedArguments:
            return 0

        newExpandedData = []
        expandedArgumentsCount = 0

        for argument in self._expandedArguments:

            extractor = CreateExtractor(argument)
            if not extractor:
                newExpandedData.append(argument)
                continue

            if not (retrievedStoryURLs := extractor.ScanChannel(argument)):

                newExpandedData.append(argument)

            else:

                newExpandedData.extend(retrievedStoryURLs)
                expandedArgumentsCount += 1

        newExpandedData = RemoveDuplicates(newExpandedData)

        self._expandedArguments = newExpandedData
        return expandedArgumentsCount

    def ExpandAndShuffle(self) -> None:

        ##
        #
        # Expands every input argument that can be expanded, recursively, then randomly reorders
        # the list.
        #
        ##

        expandedArgumentsCount = 1

        while 0 != expandedArgumentsCount:
            expandedArgumentsCount = self.Expand()

        random.shuffle(self._expandedArguments)